from __future__ import annotations

import pytest

from app.core.openai.model_registry import ReasoningLevel, UpstreamModel, get_model_registry

pytestmark = pytest.mark.integration


def _make_upstream_model(
    slug: str,
    *,
    supported_in_api: bool = True,
    base_instructions: str = "",
    raw: dict | None = None,
) -> UpstreamModel:
    return UpstreamModel(
        slug=slug,
        display_name=slug,
        description=f"Test model {slug}",
        context_window=272000,
        input_modalities=("text", "image"),
        supported_reasoning_levels=(ReasoningLevel(effort="medium", description="default"),),
        default_reasoning_level="medium",
        supports_reasoning_summaries=True,
        support_verbosity=False,
        default_verbosity=None,
        prefer_websockets=False,
        supports_parallel_tool_calls=True,
        supported_in_api=supported_in_api,
        minimal_client_version=None,
        priority=0,
        available_in_plans=frozenset({"plus", "pro"}),
        base_instructions=base_instructions,
        raw=raw
        or {
            "shell_type": "shell_command",
            "visibility": "list",
            "availability_nux": None,
        },
    )


async def _populate_test_registry() -> None:
    registry = get_model_registry()
    models = [
        _make_upstream_model("gpt-5.2"),
        _make_upstream_model("gpt-5.3-codex"),
    ]
    await registry.update({"plus": models, "pro": models})


@pytest.mark.asyncio
async def test_v1_models_list(async_client):
    await _populate_test_registry()
    resp = await async_client.get("/v1/models")
    assert resp.status_code == 200
    payload = resp.json()
    assert payload["object"] == "list"
    data = payload["data"]
    assert isinstance(data, list)
    ids = {item["id"] for item in data}
    assert "gpt-5.2" in ids
    assert "gpt-5.3-codex" in ids
    for item in data:
        assert item["object"] == "model"
        assert item["owned_by"] == "codex-lb"
        assert "metadata" in item


@pytest.mark.asyncio
async def test_v1_models_empty_when_registry_not_populated(async_client):
    registry = get_model_registry()
    registry._snapshot = None
    resp = await async_client.get("/v1/models")
    assert resp.status_code == 200
    payload = resp.json()
    assert payload["object"] == "list"
    assert payload["data"] == []


@pytest.mark.asyncio
async def test_v1_models_includes_supported_in_api_false_models(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model("gpt-5.2"),
        _make_upstream_model("gpt-5.3-codex"),
        _make_upstream_model("gpt-hidden", supported_in_api=False),
    ]
    await registry.update({"plus": models, "pro": models})

    resp = await async_client.get("/v1/models")
    assert resp.status_code == 200
    ids = {item["id"] for item in resp.json()["data"]}
    assert {"gpt-5.2", "gpt-5.3-codex", "gpt-hidden"}.issubset(ids)


@pytest.mark.asyncio
async def test_backend_codex_models_returns_format1(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model(
            "gpt-5.3-codex",
            raw={
                "shell_type": "shell_command",
                "visibility": "list",
                "availability_nux": None,
                "upgrade": {"model": "gpt-5.4", "migration_markdown": "Upgrade!"},
            },
        ),
        _make_upstream_model(
            "gpt-5.2",
            raw={
                "shell_type": "shell_command",
                "visibility": "list",
            },
        ),
    ]
    await registry.update({"plus": models, "pro": models})

    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    payload = resp.json()
    assert "models" in payload
    assert isinstance(payload["models"], list)
    slugs = {m["slug"] for m in payload["models"]}
    assert {"gpt-5.2", "gpt-5.3-codex"}.issubset(slugs)


@pytest.mark.asyncio
async def test_backend_codex_models_entry_has_upstream_fields(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model(
            "gpt-5.3-codex",
            raw={
                "shell_type": "shell_command",
                "visibility": "list",
                "availability_nux": None,
                "upgrade": {"model": "gpt-5.4", "migration_markdown": "Upgrade!"},
            },
            base_instructions="You are a helpful coding assistant.",
        ),
    ]
    await registry.update({"plus": models, "pro": models})

    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    entries = resp.json()["models"]
    entry = next(m for m in entries if m["slug"] == "gpt-5.3-codex")

    assert entry["display_name"] == "gpt-5.3-codex"
    assert entry["description"] == "Test model gpt-5.3-codex"
    assert entry["base_instructions"] == "You are a helpful coding assistant."
    assert entry["context_window"] == 272000
    assert entry["supported_in_api"] is True
    assert entry["shell_type"] == "shell_command"
    assert entry["visibility"] == "list"
    assert entry["availability_nux"] is None
    assert entry["upgrade"] == {"model": "gpt-5.4", "migration_markdown": "Upgrade!"}


@pytest.mark.asyncio
async def test_backend_codex_models_preserves_upstream_visibility(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model(
            "gpt-5.3-codex",
            raw={
                "shell_type": "shell_command",
                "visibility": "hide",
            },
        ),
    ]
    await registry.update({"plus": models, "pro": models})

    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    entries = resp.json()["models"]
    entry = next(m for m in entries if m["slug"] == "gpt-5.3-codex")
    assert entry["visibility"] == "hide"


@pytest.mark.asyncio
async def test_backend_codex_models_filters_disallowed_models(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model("gpt-5.2", base_instructions="allowed"),
        _make_upstream_model("gpt-5.3-codex", base_instructions="blocked"),
    ]
    await registry.update({"plus": models, "pro": models})

    enable = await async_client.put(
        "/api/settings",
        json={
            "stickyThreadsEnabled": False,
            "preferEarlierResetAccounts": False,
            "totpRequiredOnLogin": False,
            "apiKeyAuthEnabled": True,
        },
    )
    assert enable.status_code == 200

    created = await async_client.post(
        "/api/api-keys/",
        json={
            "name": "codex-restricted",
            "allowedModels": ["gpt-5.2"],
        },
    )
    assert created.status_code == 200
    key = created.json()["key"]

    resp = await async_client.get("/backend-api/codex/models", headers={"Authorization": f"Bearer {key}"})
    assert resp.status_code == 200
    entries = resp.json()["models"]
    assert [entry["slug"] for entry in entries] == ["gpt-5.2"]
    assert entries[0]["base_instructions"] == "allowed"


@pytest.mark.asyncio
async def test_backend_codex_models_includes_supported_in_api_false_models(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model("gpt-5.2"),
        _make_upstream_model("gpt-5.3-codex"),
        _make_upstream_model("gpt-hidden", supported_in_api=False),
    ]
    await registry.update({"plus": models, "pro": models})

    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    slugs = {m["slug"] for m in resp.json()["models"]}
    assert {"gpt-5.2", "gpt-5.3-codex", "gpt-hidden"}.issubset(slugs)


@pytest.mark.asyncio
async def test_backend_codex_models_empty_when_registry_not_populated(async_client):
    registry = get_model_registry()
    registry._snapshot = None
    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    payload = resp.json()
    assert payload["models"] == []


@pytest.mark.asyncio
async def test_model_sets_are_consistent_across_api_endpoints(async_client):
    registry = get_model_registry()
    models = [
        _make_upstream_model("gpt-5.2"),
        _make_upstream_model("gpt-5.3-codex"),
        _make_upstream_model("gpt-hidden", supported_in_api=False),
    ]
    await registry.update({"plus": models, "pro": models})

    dashboard = await async_client.get("/api/models")
    v1 = await async_client.get("/v1/models")
    codex = await async_client.get("/backend-api/codex/models")

    assert dashboard.status_code == 200
    assert v1.status_code == 200
    assert codex.status_code == 200

    dashboard_ids = {item["id"] for item in dashboard.json()["models"]}
    v1_ids = {item["id"] for item in v1.json()["data"]}
    codex_slugs = {item["slug"] for item in codex.json()["models"]}
    assert dashboard_ids == v1_ids == codex_slugs


@pytest.mark.asyncio
async def test_model_context_window_override(async_client, monkeypatch):
    registry = get_model_registry()
    models = [_make_upstream_model("gpt-5.4")]
    await registry.update({"pro": models})

    from app.core.config.settings import get_settings
    from app.modules.proxy import api as proxy_api_module

    original_settings = get_settings()
    patched = original_settings.model_copy(update={"model_context_window_overrides": {"gpt-5.4": 515000}})
    monkeypatch.setattr(proxy_api_module, "get_settings", lambda: patched)

    # /backend-api/codex/models
    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    entry = next(m for m in resp.json()["models"] if m["slug"] == "gpt-5.4")
    assert entry["context_window"] == 515000

    # /v1/models
    resp_v1 = await async_client.get("/v1/models")
    assert resp_v1.status_code == 200
    v1_entry = next(m for m in resp_v1.json()["data"] if m["id"] == "gpt-5.4")
    assert v1_entry["metadata"]["context_window"] == 515000


@pytest.mark.asyncio
async def test_model_context_window_no_override(async_client):
    registry = get_model_registry()
    models = [_make_upstream_model("gpt-5.4")]
    await registry.update({"pro": models})

    resp = await async_client.get("/backend-api/codex/models")
    assert resp.status_code == 200
    entry = next(m for m in resp.json()["models"] if m["slug"] == "gpt-5.4")
    assert entry["context_window"] == 272000



from __future__ import annotations

import logging
import sqlite3
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, AsyncIterator, Awaitable, Callable, Protocol, TypeVar

import anyio
from anyio import to_thread
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker, create_async_engine

from app.core.config.settings import get_settings
from app.db.sqlite_utils import SqliteIntegrityCheckMode, check_sqlite_integrity, sqlite_db_path_from_url

if TYPE_CHECKING:
    from app.db.migrate import MigrationRunResult, MigrationState

_settings = get_settings()

logger = logging.getLogger(__name__)

_SQLITE_BUSY_TIMEOUT_MS = 5_000
_SQLITE_BUSY_TIMEOUT_SECONDS = _SQLITE_BUSY_TIMEOUT_MS / 1000


def _is_sqlite_url(url: str) -> bool:
    return url.startswith("sqlite+aiosqlite:///") or url.startswith("sqlite:///")


def _is_sqlite_memory_url(url: str) -> bool:
    return _is_sqlite_url(url) and ":memory:" in url


def _configure_sqlite_engine(engine: Engine, *, enable_wal: bool) -> None:
    @event.listens_for(engine, "connect")
    def _set_sqlite_pragmas(dbapi_connection: sqlite3.Connection, _: object) -> None:
        cursor: sqlite3.Cursor = dbapi_connection.cursor()
        try:
            if enable_wal:
                cursor.execute("PRAGMA journal_mode=WAL")
            cursor.execute("PRAGMA synchronous=NORMAL")
            cursor.execute("PRAGMA foreign_keys=ON")
            cursor.execute(f"PRAGMA busy_timeout={_SQLITE_BUSY_TIMEOUT_MS}")
        finally:
            cursor.close()


if _is_sqlite_url(_settings.database_url):
    is_sqlite_memory = _is_sqlite_memory_url(_settings.database_url)
    if is_sqlite_memory:
        engine = create_async_engine(
            _settings.database_url,
            echo=False,
            connect_args={"timeout": _SQLITE_BUSY_TIMEOUT_SECONDS},
        )
    else:
        engine = create_async_engine(
            _settings.database_url,
            echo=False,
            pool_size=_settings.database_pool_size,
            max_overflow=_settings.database_max_overflow,
            pool_timeout=_settings.database_pool_timeout_seconds,
            connect_args={"timeout": _SQLITE_BUSY_TIMEOUT_SECONDS},
        )
    _configure_sqlite_engine(engine.sync_engine, enable_wal=not is_sqlite_memory)
else:
    engine = create_async_engine(
        _settings.database_url,
        echo=False,
        pool_size=_settings.database_pool_size,
        max_overflow=_settings.database_max_overflow,
        pool_timeout=_settings.database_pool_timeout_seconds,
    )

SessionLocal = async_sessionmaker(engine, expire_on_commit=False, class_=AsyncSession)

_background_engine: AsyncEngine | None = None
_background_session_factory: async_sessionmaker[AsyncSession] | None = None

_T = TypeVar("_T")


class _SqliteBackupCreator(Protocol):
    def __call__(self, source: Path, *, max_files: int) -> Path: ...


def _ensure_sqlite_dir(url: str) -> None:
    if not (url.startswith("sqlite+aiosqlite:") or url.startswith("sqlite:")):
        return

    marker = ":///"
    marker_index = url.find(marker)
    if marker_index < 0:
        return

    # Works for both relative (sqlite+aiosqlite:///./db.sqlite) and absolute
    # paths (sqlite+aiosqlite:////var/lib/app/db.sqlite).
    path = url[marker_index + len(marker) :]
    path = path.partition("?")[0]
    path = path.partition("#")[0]

    if not path or path == ":memory:":
        return

    Path(path).expanduser().parent.mkdir(parents=True, exist_ok=True)


def _startup_sqlite_check_mode(raw_mode: str) -> SqliteIntegrityCheckMode | None:
    if raw_mode == "off":
        return None
    return SqliteIntegrityCheckMode(raw_mode)


async def _shielded(awaitable: Awaitable[_T]) -> _T:
    with anyio.CancelScope(shield=True):
        return await awaitable


async def _safe_rollback(session: AsyncSession) -> None:
    if not session.in_transaction():
        return
    try:
        await _shielded(session.rollback())
    except BaseException:
        return


async def _safe_close(session: AsyncSession) -> None:
    try:
        await _shielded(session.close())
    except BaseException:
        return


def _load_migration_entrypoints() -> tuple[
    Callable[[str], "MigrationState"],
    Callable[[str], Awaitable["MigrationRunResult"]],
    Callable[[str], tuple[str, ...]],
]:
    from app.db.migrate import check_schema_drift, inspect_migration_state, run_startup_migrations

    return inspect_migration_state, run_startup_migrations, check_schema_drift


def _load_sqlite_backup_creator() -> _SqliteBackupCreator:
    from app.db.backup import create_sqlite_pre_migration_backup

    return create_sqlite_pre_migration_backup


def init_background_db(url: str | None = None) -> None:
    """Initialize separate DB pool for background tasks (smaller pool).

    Args:
        url: Database URL. If None, uses settings.database_url.
    """
    global _background_engine, _background_session_factory
    db_url = url or _settings.database_url

    if _is_sqlite_url(db_url):
        is_sqlite_memory = _is_sqlite_memory_url(db_url)
        if is_sqlite_memory:
            # Reuse the main engine for in-memory SQLite — creating a second
            # engine would open a separate, empty in-memory database with no
            # schema, causing "no such table" errors in background tasks.
            _background_engine = engine
            _background_session_factory = SessionLocal
            return
        else:
            _background_engine = create_async_engine(
                db_url,
                echo=False,
                pool_size=3,
                max_overflow=2,
                pool_timeout=_settings.database_pool_timeout_seconds,
                connect_args={"timeout": _SQLITE_BUSY_TIMEOUT_SECONDS},
            )
        _configure_sqlite_engine(_background_engine.sync_engine, enable_wal=not is_sqlite_memory)
    else:
        _background_engine = create_async_engine(
            db_url,
            echo=False,
            pool_size=3,
            max_overflow=2,
            pool_timeout=_settings.database_pool_timeout_seconds,
        )

    _background_session_factory = async_sessionmaker(_background_engine, expire_on_commit=False, class_=AsyncSession)


@asynccontextmanager
async def get_background_session() -> AsyncIterator[AsyncSession]:
    """Session provider for background tasks, schedulers, and auth dependencies.

    Uses the separate background pool if initialized, otherwise falls back to main pool.
    """
    factory = _background_session_factory or SessionLocal
    session = factory()
    try:
        yield session
    except BaseException:
        await _safe_rollback(session)
        raise
    finally:
        if session.in_transaction():
            await _safe_rollback(session)
        await _safe_close(session)


async def get_session() -> AsyncIterator[AsyncSession]:
    session = SessionLocal()
    try:
        yield session
    except BaseException:
        await _safe_rollback(session)
        raise
    finally:
        if session.in_transaction():
            await _safe_rollback(session)
        await _safe_close(session)


async def init_db() -> None:
    _ensure_sqlite_dir(_settings.database_url)
    sqlite_path = sqlite_db_path_from_url(_settings.database_url)
    if sqlite_path is not None:
        check_mode = _startup_sqlite_check_mode(_settings.database_sqlite_startup_check_mode)
        if check_mode is not None:
            integrity = check_sqlite_integrity(sqlite_path, mode=check_mode)
            if not integrity.ok:
                details = integrity.details or "unknown error"
                pragma_name = "quick_check" if check_mode == SqliteIntegrityCheckMode.QUICK else "integrity_check"
                logger.error(
                    "SQLite %s failed path=%s details=%s",
                    pragma_name,
                    sqlite_path,
                    details,
                )
                if "locked" in details.lower():
                    message = (
                        f"SQLite {pragma_name} failed for {sqlite_path} ({details}). "
                        "Another instance may be running. Stop it and retry."
                    )
                else:
                    message = (
                        f"SQLite {pragma_name} failed for {sqlite_path} ({details}). "
                        "The database appears corrupted or the filesystem is unhealthy. "
                        "Stop the app and run "
                        f'`python -m app.db.recover --db "{sqlite_path}" --replace` '
                        "or restore a backup from the same directory."
                    )
                raise RuntimeError(message)

    try:
        inspect_migration_state, run_startup_migrations, check_schema_drift = _load_migration_entrypoints()
    except ModuleNotFoundError as exc:
        if exc.name != "app.db.migrate":
            raise
        logger.exception("Failed to import migration entrypoint module=app.db.migrate")
        raise RuntimeError("Database migration entrypoint app.db.migrate is unavailable") from exc
    except ImportError as exc:
        logger.exception("Failed to import database migration entrypoints from app.db.migrate")
        raise RuntimeError("Database migration entrypoint app.db.migrate is invalid") from exc

    if not _settings.database_migrate_on_startup:
        migration_state = await to_thread.run_sync(
            lambda: inspect_migration_state(_settings.database_url),
        )
        if migration_state.needs_upgrade:
            current_revision = migration_state.current_revision or "none"
            message = (
                "Startup database migration is disabled but database schema is behind Alembic head "
                f"(current={current_revision}, head={migration_state.head_revision}). "
                "Run the dedicated migration job or `python -m app.db.migrate upgrade` before starting the app."
            )
            logger.error(message)
            raise RuntimeError(message)

        logger.info("Startup database migration is disabled and database schema is current")
        return

    if sqlite_path is not None and _settings.database_sqlite_pre_migrate_backup_enabled and sqlite_path.exists():
        migration_state = await to_thread.run_sync(
            lambda: inspect_migration_state(_settings.database_url),
        )
        if migration_state.needs_upgrade:
            try:
                create_sqlite_pre_migration_backup = _load_sqlite_backup_creator()
            except ModuleNotFoundError as exc:
                if exc.name != "app.db.backup":
                    raise
                logger.exception("Failed to import SQLite backup module=app.db.backup")
                raise RuntimeError("SQLite backup module app.db.backup is unavailable") from exc

            backup_path = await to_thread.run_sync(
                lambda: create_sqlite_pre_migration_backup(
                    sqlite_path,
                    max_files=_settings.database_sqlite_pre_migrate_backup_max_files,
                ),
            )
            logger.info(
                "Created SQLite pre-migration backup path=%s target_revision=%s",
                backup_path,
                migration_state.head_revision,
            )

    try:
        result = await run_startup_migrations(_settings.database_url)
        if result.bootstrap.stamped_revision is not None:
            logger.info(
                "Bootstrapped legacy migrations stamped_revision=%s legacy_rows=%s",
                result.bootstrap.stamped_revision,
                result.bootstrap.legacy_row_count,
            )
        if result.current_revision is not None:
            logger.info("Database migration complete revision=%s", result.current_revision)
        drift = await to_thread.run_sync(lambda: check_schema_drift(_settings.database_url))
        if drift:
            drift_details = "; ".join(drift)
            raise RuntimeError(f"Schema drift detected after startup migrations: {drift_details}")
    except Exception:
        logger.exception("Failed to apply database migrations")
        if _settings.database_migrations_fail_fast:
            raise


async def close_db() -> None:
    await engine.dispose()
    if _background_engine is not None:
        await _background_engine.dispose()



import asyncio
import json

from fastapi import FastAPI
from fastapi.responses import StreamingResponse

app = FastAPI()


@app.post("/backend-api/conversation")
async def mock_conversation() -> StreamingResponse:
    async def generate():
        for i in range(5):
            chunk = f"data: {json.dumps({'message': {'content': {'parts': [f'Token {i}']}}})}\n\n"
            yield chunk.encode()
            await asyncio.sleep(0.05)
        yield b"data: [DONE]\n\n"

    return StreamingResponse(generate(), media_type="text/event-stream")


@app.get("/public-api/me")
async def mock_me() -> dict[str, str]:
    return {"id": "user-123", "email": "test@example.com"}


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8080)



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

import anyio

if TYPE_CHECKING:
    from app.modules.proxy.load_balancer import SelectionInputs

_AssignedAccountsKey = tuple[str, ...] | None
_CacheKey = tuple[str | None, str | None, _AssignedAccountsKey]


@dataclass(slots=True)
class _CachedSelectionInputs:
    data: SelectionInputs
    expires_at: float


class AccountSelectionCache:
    def __init__(self, ttl_seconds: int | None = None) -> None:
        if ttl_seconds is None:
            import sys

            ttl_seconds = 0 if "pytest" in sys.modules else 5
        if ttl_seconds < 0:
            raise ValueError("ttl_seconds must be non-negative")
        self._ttl_seconds = ttl_seconds
        self._cache: dict[_CacheKey, _CachedSelectionInputs] = {}
        self._lock = anyio.Lock()
        self._generation: int = 0

    @property
    def generation(self) -> int:
        return self._generation

    async def get(self, key: _CacheKey = (None, None, None)) -> SelectionInputs | None:
        if self._ttl_seconds == 0:
            return None
        entry = self._cache.get(key)
        if entry is None:
            return None
        if time.monotonic() >= entry.expires_at:
            return None
        return entry.data

    async def set(
        self,
        data: SelectionInputs,
        key: _CacheKey = (None, None, None),
        *,
        generation: int | None = None,
    ) -> None:
        async with self._lock:
            if generation is not None and generation != self._generation:
                return
            self._cache[key] = _CachedSelectionInputs(
                data=data,
                expires_at=time.monotonic() + self._ttl_seconds,
            )

    def invalidate(self) -> None:
        self._generation += 1
        self._cache.clear()


_account_selection_cache = AccountSelectionCache()


def get_account_selection_cache() -> AccountSelectionCache:
    return _account_selection_cache



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from types import SimpleNamespace
from typing import Any, cast

import pytest

import app.modules.proxy.api as proxy_api_module
import app.modules.proxy.service as proxy_service_module
from app.core.crypto import TokenEncryptor
from app.db.models import AccountStatus, OpenAIPlatformIdentity, StickySessionKind
from app.modules.proxy.load_balancer import LoadBalancer
from app.modules.proxy.repo_bundle import ProxyRepositories
from app.modules.proxy.sticky_repository import StickyRoutingTarget
from app.modules.upstream_identities.types import (
    BACKEND_CODEX_HTTP_ROUTE_FAMILY,
    CHATGPT_PRIVATE_ROUTE_CLASS,
    OPENAI_PLATFORM_PROVIDER_KIND,
    OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
    PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
)

pytestmark = pytest.mark.unit


def _responses_request(**overrides):
    payload = {
        "model": "gpt-5.1",
        "instructions": "hi",
        "input": "hi",
    }
    payload.update(overrides)
    return proxy_service_module.ResponsesRequest.model_validate(payload)


class DummyStickyRepository:
    def __init__(self, target: StickyRoutingTarget | None = None) -> None:
        self._target = target
        self.upsert_calls: list[tuple[str, StickySessionKind, str, str]] = []
        self.delete_calls: list[tuple[str, StickySessionKind, str]] = []

    async def get_target(
        self,
        key: str,
        *,
        kind: StickySessionKind,
        provider_kind: str,
        max_age_seconds: int | None = None,
    ) -> StickyRoutingTarget | None:
        del key, kind, provider_kind, max_age_seconds
        return self._target

    async def upsert_target(
        self,
        key: str,
        *,
        kind: StickySessionKind,
        provider_kind: str,
        routing_subject_id: str,
        account_id: str | None = None,
    ):
        del account_id
        self.upsert_calls.append((key, kind, provider_kind, routing_subject_id))
        return SimpleNamespace()

    async def delete_scoped(self, key: str, *, kind: StickySessionKind, provider_kind: str) -> bool:
        self.delete_calls.append((key, kind, provider_kind))
        return True


class DummyPlatformIdentitiesRepository:
    def __init__(self, identities: list[OpenAIPlatformIdentity]) -> None:
        self._identities = identities

    async def list_eligible_identities(self, route_family: str) -> list[OpenAIPlatformIdentity]:
        return [
            identity
            for identity in self._identities
            if route_family in (identity.eligible_route_families or "")
            and identity.status not in (AccountStatus.PAUSED, AccountStatus.DEACTIVATED)
        ]

    async def get_by_id(self, identity_id: str) -> OpenAIPlatformIdentity | None:
        return next((identity for identity in self._identities if identity.id == identity_id), None)


def _platform_identity(identity_id: str) -> OpenAIPlatformIdentity:
    return OpenAIPlatformIdentity(
        id=identity_id,
        label=f"Platform {identity_id}",
        api_key_encrypted=TokenEncryptor().encrypt(f"sk-{identity_id}"),
        organization_id="org_test",
        project_id="proj_test",
        eligible_route_families="public_models_http,public_responses_http,backend_codex_http",
        status=AccountStatus.ACTIVE,
        last_validated_at=None,
        last_auth_failure_reason=None,
        deactivation_reason=None,
    )


@asynccontextmanager
async def _repo_factory(
    *,
    identities: list[OpenAIPlatformIdentity] | None = None,
    sticky_repo: DummyStickyRepository | None = None,
) -> AsyncIterator[ProxyRepositories]:
    yield cast(
        ProxyRepositories,
        SimpleNamespace(
            accounts=SimpleNamespace(),
            platform_identities=DummyPlatformIdentitiesRepository(identities or []),
            usage=SimpleNamespace(),
            request_logs=SimpleNamespace(),
            sticky_sessions=sticky_repo or DummyStickyRepository(),
            api_keys=SimpleNamespace(),
            additional_usage=SimpleNamespace(),
        ),
    )


def test_derive_request_capabilities_marks_stateless_public_http_as_platform_eligible() -> None:
    capabilities = proxy_api_module._derive_request_capabilities(
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
        transport="http",
        model="gpt-5.1",
        payload=_responses_request(),
        headers={},
    )

    assert capabilities.route_family == PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY
    assert capabilities.transport == "http"
    assert capabilities.continuity_param is None


def test_derive_request_capabilities_marks_previous_response_id_as_continuity() -> None:
    capabilities = proxy_api_module._derive_request_capabilities(
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
        transport="http",
        model="gpt-5.1",
        payload=_responses_request(previous_response_id="resp_prev_1"),
        headers={},
    )

    assert capabilities.continuity_param == "previous_response_id"


@pytest.mark.parametrize(
    "header_name",
    ["session_id", "x-codex-session-id", "x-codex-conversation-id", "x-codex-turn-state"],
)
def test_derive_request_capabilities_marks_session_headers_as_continuity(header_name: str) -> None:
    capabilities = proxy_api_module._derive_request_capabilities(
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
        transport="http",
        model="gpt-5.1",
        payload=_responses_request(),
        headers={header_name: "sid_1"},
    )

    assert capabilities.continuity_param == header_name


@pytest.mark.parametrize(
    "header_name",
    ["session_id", "x-codex-session-id", "x-codex-conversation-id", "x-codex-turn-state"],
)
def test_derive_request_capabilities_marks_backend_codex_session_headers_as_continuity(
    header_name: str,
) -> None:
    capabilities = proxy_api_module._derive_request_capabilities(
        route_family=BACKEND_CODEX_HTTP_ROUTE_FAMILY,
        route_class=CHATGPT_PRIVATE_ROUTE_CLASS,
        transport="http",
        model="gpt-5.1",
        payload=_responses_request(),
        headers={header_name: "sid_1"},
    )

    assert capabilities.continuity_param == header_name


@pytest.mark.parametrize(
    "header_name",
    ["session_id", "x-codex-session-id", "x-codex-conversation-id", "x-codex-turn-state"],
)
def test_derive_request_capabilities_marks_backend_codex_header_continuity(header_name: str) -> None:
    capabilities = proxy_api_module._derive_request_capabilities(
        route_family=BACKEND_CODEX_HTTP_ROUTE_FAMILY,
        route_class=CHATGPT_PRIVATE_ROUTE_CLASS,
        transport="http",
        model="gpt-5.1",
        payload=_responses_request(),
        headers={header_name: "sid_1"},
    )

    assert capabilities.continuity_param == header_name


@pytest.mark.asyncio
async def test_select_routing_subject_keeps_chatgpt_primary_when_public_http_is_healthy(monkeypatch) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())
    identity = proxy_service_module._SelectedPlatformIdentity(
        id="plat_1",
        api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
        organization_id="org_test",
        project_id="proj_test",
    )

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_should_fallback(*, model: str | None, account_ids=None) -> bool:
        del model, account_ids
        return False

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return identity

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "should_fallback_to_platform_for_usage_drain", fake_should_fallback)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
        )
    )

    assert result.is_chatgpt is True
    selected = result.selected
    assert isinstance(selected, proxy_service_module.SelectedChatGPTSubject)
    assert selected.provider_kind == "chatgpt_web"


@pytest.mark.asyncio
async def test_select_routing_subject_uses_platform_when_public_http_usage_is_drained(monkeypatch) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())
    identity = proxy_service_module._SelectedPlatformIdentity(
        id="plat_1",
        api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
        organization_id="org_test",
        project_id="proj_test",
    )

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_should_fallback(*, model: str | None, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return identity

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "should_fallback_to_platform_for_usage_drain", fake_should_fallback)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
        )
    )

    assert result.is_platform is True
    selected = result.selected
    assert isinstance(selected, proxy_service_module.SelectedPlatformSubject)
    assert selected.provider_kind == OPENAI_PLATFORM_PROVIDER_KIND
    assert selected.routing_subject_id == "plat_1"


@pytest.mark.asyncio
async def test_select_routing_subject_uses_platform_for_models_when_public_http_usage_is_drained(
    monkeypatch,
) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())
    identity = proxy_service_module._SelectedPlatformIdentity(
        id="plat_1",
        api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
        organization_id="org_test",
        project_id="proj_test",
    )

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_should_fallback(*, model: str | None, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return identity

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "should_fallback_to_platform_for_usage_drain", fake_should_fallback)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family="public_models_http",
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model=None,
        )
    )

    assert result.is_platform is True
    selected = result.selected
    assert isinstance(selected, proxy_service_module.SelectedPlatformSubject)
    assert selected.provider_kind == OPENAI_PLATFORM_PROVIDER_KIND
    assert selected.routing_subject_id == "plat_1"


@pytest.mark.asyncio
async def test_select_routing_subject_uses_platform_for_backend_codex_http_when_usage_is_drained(
    monkeypatch,
) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())
    identity = proxy_service_module._SelectedPlatformIdentity(
        id="plat_1",
        api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
        organization_id="org_test",
        project_id="proj_test",
    )

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_should_fallback(*, model: str | None, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del kwargs
        assert route_family == BACKEND_CODEX_HTTP_ROUTE_FAMILY
        return identity

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "should_fallback_to_platform_for_usage_drain", fake_should_fallback)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=BACKEND_CODEX_HTTP_ROUTE_FAMILY,
            route_class=CHATGPT_PRIVATE_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
        )
    )

    assert result.is_platform is True
    selected = result.selected
    assert isinstance(selected, proxy_service_module.SelectedPlatformSubject)
    assert selected.provider_kind == OPENAI_PLATFORM_PROVIDER_KIND
    assert selected.routing_subject_id == "plat_1"


@pytest.mark.asyncio
async def test_select_routing_subject_falls_back_to_chatgpt_for_continuity(monkeypatch) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return True

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return proxy_service_module._SelectedPlatformIdentity(
            id="plat_1",
            api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
            organization_id=None,
            project_id=None,
        )

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
            continuity_param="previous_response_id",
        )
    )

    assert result.is_chatgpt is True
    selected = result.selected
    assert isinstance(selected, proxy_service_module.SelectedChatGPTSubject)
    assert selected.provider_kind == "chatgpt_web"


@pytest.mark.asyncio
async def test_select_routing_subject_returns_provider_continuity_failure_when_only_platform(monkeypatch) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return False

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return proxy_service_module._SelectedPlatformIdentity(
            id="plat_1",
            api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
            organization_id=None,
            project_id=None,
        )

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
            continuity_param="previous_response_id",
        )
    )

    assert result.failure is not None
    assert result.failure.error_code == "provider_continuity_unsupported"
    assert result.failure.error_param == "previous_response_id"


@pytest.mark.asyncio
async def test_select_routing_subject_does_not_use_platform_when_only_platform_exists(monkeypatch) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return False

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del route_family, kwargs
        return proxy_service_module._SelectedPlatformIdentity(
            id="plat_1",
            api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
            organization_id=None,
            project_id=None,
        )

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
            route_class=OPENAI_PUBLIC_HTTP_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
        )
    )

    assert result.selected is None
    assert result.failure is not None
    assert result.failure.error_code == "provider_fallback_requires_chatgpt"


@pytest.mark.asyncio
async def test_select_routing_subject_requires_chatgpt_for_backend_codex_http_when_only_platform_exists(
    monkeypatch,
) -> None:
    service = proxy_service_module.ProxyService(lambda: _repo_factory())

    async def fake_has_chatgpt_candidates(model: str | None = None, *, account_ids=None) -> bool:
        del model, account_ids
        return False

    async def fake_select_platform_identity(route_family: str, **kwargs):
        del kwargs
        assert route_family == BACKEND_CODEX_HTTP_ROUTE_FAMILY
        return proxy_service_module._SelectedPlatformIdentity(
            id="plat_1",
            api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
            organization_id=None,
            project_id=None,
        )

    monkeypatch.setattr(service, "has_chatgpt_candidates", fake_has_chatgpt_candidates)
    monkeypatch.setattr(service, "select_platform_identity", fake_select_platform_identity)

    result = await service.select_routing_subject(
        capabilities=proxy_service_module.RequestCapabilities(
            route_family=BACKEND_CODEX_HTTP_ROUTE_FAMILY,
            route_class=CHATGPT_PRIVATE_ROUTE_CLASS,
            transport="http",
            model="gpt-5.1",
        )
    )

    assert result.selected is None
    assert result.failure is not None
    assert result.failure.error_code == "provider_fallback_requires_chatgpt"


@pytest.mark.asyncio
async def test_load_balancer_select_routing_subject_uses_platform_prompt_cache_sticky() -> None:
    sticky_repo = DummyStickyRepository(
        StickyRoutingTarget(
            provider_kind=OPENAI_PLATFORM_PROVIDER_KIND,
            routing_subject_id="plat_2",
            account_id=None,
        )
    )
    identities = [_platform_identity("plat_1"), _platform_identity("plat_2")]
    balancer = LoadBalancer(lambda: _repo_factory(identities=identities, sticky_repo=sticky_repo))

    result = await balancer.select_routing_subject(
        provider_kind=OPENAI_PLATFORM_PROVIDER_KIND,
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        sticky_key="cache-key",
        sticky_kind=StickySessionKind.PROMPT_CACHE,
        sticky_max_age_seconds=300,
    )

    assert result.provider_kind == OPENAI_PLATFORM_PROVIDER_KIND
    assert result.routing_subject_id == "plat_2"
    assert sticky_repo.upsert_calls == [
        ("cache-key", StickySessionKind.PROMPT_CACHE, OPENAI_PLATFORM_PROVIDER_KIND, "plat_2")
    ]


@pytest.mark.asyncio
async def test_load_balancer_select_routing_subject_discards_stale_platform_sticky_target() -> None:
    sticky_repo = DummyStickyRepository(
        StickyRoutingTarget(
            provider_kind=OPENAI_PLATFORM_PROVIDER_KIND,
            routing_subject_id="plat_missing",
            account_id=None,
        )
    )
    balancer = LoadBalancer(lambda: _repo_factory(identities=[_platform_identity("plat_1")], sticky_repo=sticky_repo))

    result = await balancer.select_routing_subject(
        provider_kind=OPENAI_PLATFORM_PROVIDER_KIND,
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        sticky_key="cache-key",
        sticky_kind=StickySessionKind.PROMPT_CACHE,
    )

    assert result.provider_kind == OPENAI_PLATFORM_PROVIDER_KIND
    assert result.routing_subject_id == "plat_1"
    assert sticky_repo.delete_calls == [("cache-key", StickySessionKind.PROMPT_CACHE, OPENAI_PLATFORM_PROVIDER_KIND)]


@pytest.mark.asyncio
async def test_platform_only_route_rejection_respects_api_key_account_scope() -> None:
    seen: dict[str, object] = {}

    class DummyService:
        async def select_platform_identity(self, route_family: str):
            seen["route_family"] = route_family
            return proxy_service_module._SelectedPlatformIdentity(
                id="plat_1",
                api_key_encrypted=TokenEncryptor().encrypt("sk-platform"),
                organization_id=None,
                project_id=None,
            )

        async def has_chatgpt_candidates(self, model: str | None, *, account_ids=None) -> bool:
            seen["model"] = model
            seen["account_ids"] = list(account_ids) if account_ids is not None else None
            return False

    api_key = SimpleNamespace(
        account_assignment_scope_enabled=True,
        assigned_account_ids=["acc_scoped"],
    )
    result = await proxy_api_module._should_reject_platform_only_route(
        context=cast(Any, SimpleNamespace(service=DummyService())),
        api_key=cast(Any, api_key),
        route_family=PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        model="gpt-5.1",
    )

    assert result is True
    assert seen == {
        "route_family": PUBLIC_RESPONSES_HTTP_ROUTE_FAMILY,
        "model": "gpt-5.1",
        "account_ids": ["acc_scoped"],
    }



import os
import subprocess
import sys


def run(cmd: list[str]) -> None:
    subprocess.run(cmd, check=True)


if __name__ == "__main__":
    run([sys.executable, "-m", "app.db.migrate", "upgrade"])
    os.environ["CODEX_LB_DATABASE_MIGRATE_ON_STARTUP"] = "false"
    os.execvp("fastapi", ["fastapi", "run", "app/main.py", "--host", "0.0.0.0", "--port", "2455"])



from __future__ import annotations

import logging

from fastapi import FastAPI, Request
from fastapi.exception_handlers import (
    http_exception_handler,
    request_validation_exception_handler,
)
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse, Response
from starlette.exceptions import HTTPException as StarletteHTTPException

from app.core.errors import dashboard_error, openai_error
from app.core.exceptions import (
    AppError,
    DashboardAuthError,
    DashboardBadRequestError,
    DashboardConflictError,
    DashboardNotFoundError,
    DashboardRateLimitError,
    DashboardValidationError,
    ProxyAuthError,
    ProxyModelNotAllowed,
    ProxyRateLimitError,
    ProxyUpstreamError,
)
from app.core.runtime_logging import log_error_response

logger = logging.getLogger(__name__)

_OPENAI_EXCEPTION_TYPES: tuple[type[AppError], ...] = (
    ProxyAuthError,
    ProxyModelNotAllowed,
    ProxyRateLimitError,
    ProxyUpstreamError,
)

_DASHBOARD_EXCEPTION_TYPES: tuple[type[AppError], ...] = (
    DashboardAuthError,
    DashboardNotFoundError,
    DashboardConflictError,
    DashboardBadRequestError,
    DashboardValidationError,
    DashboardRateLimitError,
)


def _error_format(request: Request) -> str | None:
    fmt = getattr(request.state, "error_format", None)
    if fmt is not None:
        return fmt
    # Fallback for unmatched routes (e.g. SPA fallback 404s)
    path = request.url.path
    if path.startswith("/api/"):
        return "dashboard"
    if path.startswith("/v1/") or path.startswith("/backend-api/"):
        return "openai"
    return None


def add_exception_handlers(app: FastAPI) -> None:
    # --- Domain exceptions: OpenAI envelope ---

    for exc_cls in _OPENAI_EXCEPTION_TYPES:

        @app.exception_handler(exc_cls)
        async def _openai_domain_handler(request: Request, exc: AppError) -> JSONResponse:
            error_type = getattr(exc, "error_type", "server_error")
            log_error_response(
                logger,
                request,
                exc.status_code,
                exc.code,
                exc.message,
                category="openai_error_response",
            )
            return JSONResponse(
                status_code=exc.status_code,
                content=openai_error(exc.code, exc.message, error_type=error_type),
            )

    # --- Domain exceptions: Dashboard envelope ---

    for exc_cls in _DASHBOARD_EXCEPTION_TYPES:

        @app.exception_handler(exc_cls)
        async def _dashboard_domain_handler(request: Request, exc: AppError) -> JSONResponse:
            headers: dict[str, str] | None = None
            if isinstance(exc, DashboardRateLimitError):
                headers = {"Retry-After": str(exc.retry_after)}
            log_error_response(
                logger,
                request,
                exc.status_code,
                exc.code,
                exc.message,
                category="dashboard_error_response",
            )
            return JSONResponse(
                status_code=exc.status_code,
                content=dashboard_error(exc.code, exc.message),
                headers=headers,
            )

    # --- Framework exceptions: format based on router marker ---

    @app.exception_handler(RequestValidationError)
    async def validation_error_handler(
        request: Request,
        exc: RequestValidationError,
    ) -> Response:
        first_message: str | None = None
        if exc.errors():
            first = exc.errors()[0]
            message = first.get("msg")
            if isinstance(message, str):
                first_message = message
        fmt = _error_format(request)
        if fmt == "dashboard":
            log_error_response(
                logger,
                request,
                422,
                "validation_error",
                first_message or "Invalid request payload",
                category="dashboard_error_response",
            )
            return JSONResponse(
                status_code=422,
                content=dashboard_error("validation_error", "Invalid request payload"),
            )
        if fmt == "openai":
            error = openai_error("invalid_request_error", "Invalid request payload", error_type="invalid_request_error")
            if exc.errors():
                first = exc.errors()[0]
                loc = first.get("loc", [])
                if isinstance(loc, (list, tuple)):
                    param = ".".join(str(part) for part in loc if part != "body")
                    if param:
                        error["error"]["param"] = param
            log_error_response(
                logger,
                request,
                400,
                "invalid_request_error",
                first_message or "Invalid request payload",
                category="openai_error_response",
            )
            return JSONResponse(status_code=400, content=error)
        return await request_validation_exception_handler(request, exc)

    @app.exception_handler(StarletteHTTPException)
    async def http_error_handler(
        request: Request,
        exc: StarletteHTTPException,
    ) -> Response:
        fmt = _error_format(request)
        detail = exc.detail if isinstance(exc.detail, str) else "Request failed"
        if fmt == "dashboard":
            log_error_response(
                logger,
                request,
                exc.status_code,
                f"http_{exc.status_code}",
                detail,
                category="dashboard_error_response",
            )
            return JSONResponse(
                status_code=exc.status_code,
                content=dashboard_error(f"http_{exc.status_code}", detail),
            )
        if fmt == "openai":
            error_type = "invalid_request_error"
            code = "invalid_request_error"
            if exc.status_code == 401:
                error_type = "authentication_error"
                code = "invalid_api_key"
            elif exc.status_code == 403:
                error_type = "permission_error"
                code = "insufficient_permissions"
            elif exc.status_code == 404:
                error_type = "invalid_request_error"
                code = "not_found"
            elif exc.status_code == 429:
                error_type = "rate_limit_error"
                code = "rate_limit_exceeded"
            elif exc.status_code >= 500:
                error_type = "server_error"
                code = "server_error"
            log_error_response(
                logger,
                request,
                exc.status_code,
                code,
                detail,
                category="openai_error_response",
            )
            return JSONResponse(status_code=exc.status_code, content=openai_error(code, detail, error_type=error_type))
        return await http_exception_handler(request, exc)

    # --- Catch-all for unhandled exceptions ---

    @app.exception_handler(Exception)
    async def unhandled_error_handler(request: Request, exc: Exception) -> JSONResponse:
        fmt = _error_format(request)
        category = "unhandled_error_response"
        code = "server_error"
        message = str(exc) or "Unexpected error"
        log_error_response(logger, request, 500, code, message, category=category, exc_info=True)
        if fmt == "dashboard":
            return JSONResponse(
                status_code=500,
                content=dashboard_error("internal_error", "Unexpected error"),
            )
        if fmt == "openai":
            return JSONResponse(
                status_code=500,
                content=openai_error("server_error", "Internal server error", error_type="server_error"),
            )
        return JSONResponse(status_code=500, content={"detail": "Internal Server Error"})



from __future__ import annotations

import sqlalchemy as sa
from alembic import op

revision = "20260328_000000_add_rate_limit_attempts"
down_revision = "20260325_000000_add_request_log_cost"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("rate_limit_attempts"):
        op.create_table(
            "rate_limit_attempts",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("key", sa.String(length=255), nullable=False),
            sa.Column("attempted_at", sa.DateTime(timezone=True), nullable=False),
            sa.Column("type", sa.String(length=50), nullable=False),
            sa.PrimaryKeyConstraint("id"),
        )

    index_names = {index["name"] for index in inspector.get_indexes("rate_limit_attempts")}
    if "ix_rate_limit_attempts_key" not in index_names:
        op.create_index("ix_rate_limit_attempts_key", "rate_limit_attempts", ["key"], unique=False)
    if "ix_rate_limit_attempts_type_key_attempted_at" not in index_names:
        op.create_index(
            "ix_rate_limit_attempts_type_key_attempted_at",
            "rate_limit_attempts",
            ["type", "key", "attempted_at"],
            unique=False,
        )


def downgrade() -> None:
    bind = op.get_bind()
    inspector = sa.inspect(bind)
    if not inspector.has_table("rate_limit_attempts"):
        return

    index_names = {index["name"] for index in inspector.get_indexes("rate_limit_attempts")}
    if "ix_rate_limit_attempts_type_key_attempted_at" in index_names:
        op.drop_index("ix_rate_limit_attempts_type_key_attempted_at", table_name="rate_limit_attempts")
    if "ix_rate_limit_attempts_key" in index_names:
        op.drop_index("ix_rate_limit_attempts_key", table_name="rate_limit_attempts")
    op.drop_table("rate_limit_attempts")



from __future__ import annotations

import asyncio
import importlib
from datetime import UTC, datetime, timedelta
from types import SimpleNamespace
from typing import TypedDict

import pytest

leader_election_module = importlib.import_module("app.core.scheduling.leader_election")

pytestmark = pytest.mark.unit


@pytest.mark.asyncio
async def test_try_acquire_returns_true_when_disabled(monkeypatch: pytest.MonkeyPatch) -> None:
    settings = SimpleNamespace(
        leader_election_enabled=False,
        database_url="postgresql+asyncpg://db",
        leader_election_ttl_seconds=30,
    )
    monkeypatch.setattr(leader_election_module, "get_settings", lambda: settings)

    election = leader_election_module.LeaderElection(leader_id="node-a")

    assert await election.try_acquire() is True


@pytest.mark.asyncio
async def test_try_acquire_returns_true_for_sqlite(monkeypatch: pytest.MonkeyPatch) -> None:
    settings = SimpleNamespace(
        leader_election_enabled=True,
        database_url="sqlite+aiosqlite:///tmp/test.db",
        leader_election_ttl_seconds=30,
    )
    monkeypatch.setattr(leader_election_module, "get_settings", lambda: settings)

    election = leader_election_module.LeaderElection(leader_id="node-a")

    assert await election.try_acquire() is True


class _LeaseSession:
    def __init__(self, shared: "_SharedLease", lock: asyncio.Lock) -> None:
        self._shared = shared
        self._lock = lock

    async def execute(self, _statement: object, params: dict[str, object]) -> object:
        async with self._lock:
            row = self._shared.row
            now_obj = params["now"]
            leader_id_obj = params["leader_id"]
            expires_at_obj = params["expires_at"]

            assert isinstance(now_obj, datetime)
            assert isinstance(leader_id_obj, str)
            assert isinstance(expires_at_obj, datetime)

            now = now_obj
            leader_id = leader_id_obj
            expires_at = expires_at_obj

            if row is None:
                new_row: _LeaseRow = {"leader_id": leader_id, "expires_at": expires_at}
                self._shared.row = new_row
            else:
                expired = row["expires_at"] < now
                same_leader = row["leader_id"] == leader_id
                if expired or same_leader:
                    row["leader_id"] = leader_id
                    row["expires_at"] = expires_at

        return object()

    async def commit(self) -> None:
        return None

    async def scalar(self, _statement: object) -> str | None:
        row = self._shared.row
        if row is None:
            return None
        return row["leader_id"]


class _LeaseRow(TypedDict):
    leader_id: str
    expires_at: datetime


class _SharedLease:
    def __init__(self, row: _LeaseRow | None = None) -> None:
        self.row = row


def _build_session_provider(shared: _SharedLease, lock: asyncio.Lock):
    async def _provider():
        yield _LeaseSession(shared, lock)

    return _provider


@pytest.mark.asyncio
async def test_try_acquire_concurrent_only_one_wins(monkeypatch: pytest.MonkeyPatch) -> None:
    settings = SimpleNamespace(
        leader_election_enabled=True,
        database_url="postgresql+asyncpg://db",
        leader_election_ttl_seconds=30,
    )
    monkeypatch.setattr(leader_election_module, "get_settings", lambda: settings)

    shared = _SharedLease()
    lock = asyncio.Lock()
    monkeypatch.setattr(leader_election_module, "get_session", _build_session_provider(shared, lock))

    election_a = leader_election_module.LeaderElection(leader_id="node-a")
    election_b = leader_election_module.LeaderElection(leader_id="node-b")

    result_a, result_b = await asyncio.gather(election_a.try_acquire(), election_b.try_acquire())

    assert (result_a, result_b).count(True) == 1
    assert (result_a, result_b).count(False) == 1


@pytest.mark.asyncio
async def test_try_acquire_sets_is_leader_on_success(monkeypatch: pytest.MonkeyPatch) -> None:
    settings = SimpleNamespace(
        leader_election_enabled=True,
        database_url="postgresql+asyncpg://db",
        leader_election_ttl_seconds=30,
    )
    monkeypatch.setattr(leader_election_module, "get_settings", lambda: settings)

    now = datetime.now(UTC)
    shared = _SharedLease(row={"leader_id": "node-a", "expires_at": now + timedelta(seconds=30)})
    lock = asyncio.Lock()
    monkeypatch.setattr(leader_election_module, "get_session", _build_session_provider(shared, lock))

    election = leader_election_module.LeaderElection(leader_id="node-a")

    assert await election.try_acquire() is True
    assert election._is_leader is True



from __future__ import annotations

import shutil
from datetime import datetime, timezone
from pathlib import Path


def _backup_name(source: Path, timestamp: datetime) -> str:
    return f"{source.stem}.pre-migrate-{timestamp.strftime('%Y%m%dT%H%M%SZ')}{source.suffix}"


def _next_backup_path(source: Path, timestamp: datetime) -> Path:
    base_name = _backup_name(source, timestamp)
    candidate = source.parent / base_name
    if not candidate.exists():
        return candidate

    sequence = 1
    while True:
        name = f"{source.stem}.pre-migrate-{timestamp.strftime('%Y%m%dT%H%M%SZ')}-{sequence}{source.suffix}"
        candidate = source.parent / name
        if not candidate.exists():
            return candidate
        sequence += 1


def list_sqlite_pre_migration_backups(source: Path) -> list[Path]:
    pattern = f"{source.stem}.pre-migrate-*{source.suffix}"
    return sorted((path for path in source.parent.glob(pattern) if path.is_file()))


def create_sqlite_pre_migration_backup(
    source: Path,
    *,
    max_files: int,
    now: datetime | None = None,
) -> Path:
    if max_files < 1:
        raise ValueError("max_files must be >= 1")
    if not source.exists():
        raise FileNotFoundError(f"sqlite database not found: {source}")

    timestamp = now or datetime.now(timezone.utc)
    backup_path = _next_backup_path(source, timestamp)

    shutil.copy2(source, backup_path)

    backups = list_sqlite_pre_migration_backups(source)
    excess = len(backups) - max_files
    if excess > 0:
        for old_backup in backups[:excess]:
            old_backup.unlink()

    return backup_path



from __future__ import annotations

from datetime import timedelta

from app.core import usage as usage_core
from app.core.usage.types import UsageWindowRow
from app.core.utils.time import utcnow
from app.modules.accounts.repository import AccountsRepository
from app.modules.request_logs.repository import RequestLogsRepository
from app.modules.usage.builders import (
    build_usage_history_response,
    build_usage_summary_response,
    build_usage_window_response,
)
from app.modules.usage.repository import UsageRepository
from app.modules.usage.schemas import (
    UsageHistoryResponse,
    UsageSummaryResponse,
    UsageWindowResponse,
)


class UsageService:
    def __init__(
        self,
        usage_repo: UsageRepository,
        logs_repo: RequestLogsRepository,
        accounts_repo: AccountsRepository,
    ) -> None:
        self._usage_repo = usage_repo
        self._logs_repo = logs_repo
        self._accounts_repo = accounts_repo

    async def get_usage_summary(self) -> UsageSummaryResponse:
        now = utcnow()
        accounts = await self._accounts_repo.list_accounts()

        primary_rows_raw = await self._latest_usage_rows("primary")
        secondary_rows_raw = await self._latest_usage_rows("secondary")
        primary_rows, secondary_rows = usage_core.normalize_weekly_only_rows(
            primary_rows_raw,
            secondary_rows_raw,
        )

        secondary_minutes = usage_core.resolve_window_minutes("secondary", secondary_rows)
        logs_secondary = []
        if secondary_minutes:
            logs_secondary = await self._logs_repo.list_since(now - timedelta(minutes=secondary_minutes))
        return build_usage_summary_response(
            accounts=accounts,
            primary_rows=primary_rows,
            secondary_rows=secondary_rows,
            logs_secondary=logs_secondary,
        )

    async def get_usage_history(self, hours: int) -> UsageHistoryResponse:
        now = utcnow()
        since = now - timedelta(hours=hours)
        accounts = await self._accounts_repo.list_accounts()
        usage_rows = [row.to_window_row() for row in await self._usage_repo.aggregate_since(since, window="primary")]

        return build_usage_history_response(
            hours=hours,
            usage_rows=usage_rows,
            accounts=accounts,
            window="primary",
        )

    async def get_usage_window(self, window: str) -> UsageWindowResponse:
        window_key = (window or "").lower()
        if window_key not in {"primary", "secondary"}:
            raise ValueError("window must be 'primary' or 'secondary'")
        accounts = await self._accounts_repo.list_accounts()
        primary_rows_raw = await self._latest_usage_rows("primary")
        secondary_rows_raw = await self._latest_usage_rows("secondary")
        primary_rows, secondary_rows = usage_core.normalize_weekly_only_rows(
            primary_rows_raw,
            secondary_rows_raw,
        )
        usage_rows = primary_rows if window_key == "primary" else secondary_rows
        window_minutes = usage_core.resolve_window_minutes(window_key, usage_rows)
        return build_usage_window_response(
            window_key=window_key,
            window_minutes=window_minutes,
            usage_rows=usage_rows,
            accounts=accounts,
        )

    async def _latest_usage_rows(self, window: str) -> list[UsageWindowRow]:
        latest = await self._usage_repo.latest_by_account(window=window)
        return [
            UsageWindowRow(
                account_id=entry.account_id,
                used_percent=entry.used_percent,
                reset_at=entry.reset_at,
                window_minutes=entry.window_minutes,
                recorded_at=entry.recorded_at,
            )
            for entry in latest.values()
        ]



from app.core.middleware.api_firewall import add_api_firewall_middleware
from app.core.middleware.request_decompression import add_request_decompression_middleware
from app.core.middleware.request_id import add_request_id_middleware

__all__ = [
    "add_api_firewall_middleware",
    "add_request_decompression_middleware",
    "add_request_id_middleware",
]
